/*
 *  GRUB Utilities --  Utilities for GRUB Legacy, GRUB2 and GRUB for DOS
 *  Copyright(C) 2007 Bean(bean123ch@gmail.com)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <windows.h>
#include <string.h>
#include <winioctl.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <richedit.h>
#include <windowsx.h>

#include "resource.h"
#include "version.h"
#include "utils.h"
#include "xdio.h"

HINSTANCE hInst;
PCHAR lng_ext;
char* str_tab[IDS_COUNT];
int hdn[MAX_DISKS];
char hdl[MAX_DISKS][128];

typedef enum _STORAGE_BUS_TYPE_OVERRIDE { 
  BusTypeUnknownOverride            = 0x00,
  BusTypeScsiOverride               = 0x1,
  BusTypeAtapiOverride              = 0x2,
  BusTypeAtaOverride                = 0x3,
  BusType1394Override               = 0x4,
  BusTypeSsaOverride                = 0x5,
  BusTypeFibreOverride              = 0x6,
  BusTypeUsbOverride                = 0x7,
  BusTypeRAIDOverride               = 0x8,
  BusTypeiScsiOverride              = 0x9,
  BusTypeSasOverride                = 0xA,
  BusTypeSataOverride               = 0xB,
  BusTypeSdOverride                 = 0xC,
  BusTypeMmcOverride                = 0xD,
  BusTypeVirtualOverride            = 0xE,
  BusTypeFileBackedVirtualOverride  = 0xF,
  BusTypeSpacesOverride             = 0x10,
  BusTypeMaxOverride                = 0x11,
  BusTypeMaxReservedOverride        = 0x7F
} STORAGE_BUS_TYPE_OVERRIDE, *PSTORAGE_BUS_TYPE_OVERRIDE;

typedef enum _STORAGE_PROPERTY_ID {
  StorageDeviceProperty=0,
  StorageAdapterProperty,
  StorageDeviceIdProperty
} STORAGE_PROPERTY_ID;
typedef enum _STORAGE_QUERY_TYPE {
  PropertyStandardQuery=0,
  PropertyExistsQuery,
  PropertyMaskQuery,
  PropertyQueryMaxDefined
} STORAGE_QUERY_TYPE;
     typedef struct _STORAGE_PROPERTY_QUERY {
          STORAGE_PROPERTY_ID  PropertyId;
          STORAGE_QUERY_TYPE  QueryType;
          UCHAR  AdditionalParameters[1];
     } STORAGE_PROPERTY_QUERY;
typedef struct _STORAGE_DEVICE_DESCRIPTOR {
  ULONG  Version;
  ULONG  Size;
  UCHAR  DeviceType;
  UCHAR  DeviceTypeModifier;
  BOOLEAN  RemovableMedia;
  BOOLEAN  CommandQueueing;
  ULONG  VendorIdOffset;
  ULONG  ProductIdOffset;
  ULONG  ProductRevisionOffset;
  ULONG  SerialNumberOffset;
  STORAGE_BUS_TYPE_OVERRIDE  BusType;
  ULONG  RawPropertiesLength;
  UCHAR  RawDeviceProperties[1];
} STORAGE_DEVICE_DESCRIPTOR, *PSTORAGE_DEVICE_DESCRIPTOR;
#define IOCTL_STORAGE_QUERY_PROPERTY   CTL_CODE(IOCTL_STORAGE_BASE, 0x0500, METHOD_BUFFERED, FILE_ANY_ACCESS)

void ChangeText(HWND hWnd)
{
  char LocaleName[4],buf[MAX_PATH],*pc;
  FILE *in;

  if(!lng_ext)
    {
      GetLocaleInfo(GetUserDefaultLCID(), LOCALE_SABBREVLANGNAME, LocaleName, sizeof(LocaleName));
      lng_ext=LocaleName;
    }
  GetModuleFileName(hInst,buf,sizeof(buf));
  pc=strrchr(buf,'.');
  if(pc)
    strcpy(pc+1,lng_ext);
  in=fopen(buf,"rt");
  if(in==NULL)
    return;
  while(fgets(buf,sizeof(buf),in))
    {
      char *pb;
      int nn;

      if(buf[0]=='#')
        continue;
      pb=&buf[strlen(buf)];
      while((pb!=buf)&&((*(pb-1)=='\r')||(*(pb-1)=='\n')))
        pb--;
      if(pb==buf)
        continue;
      *pb=0;
      pb=strchr(buf,'=');
      if(pb==NULL)
        continue;
      *(pb++)=0;
      nn=atoi(buf);
      if((nn>=IDS_BEGIN)&&(nn<=IDS_END))
        {
          nn-=IDS_BEGIN;
          if(str_tab[nn])
            free(str_tab[nn]);
          str_tab[nn]=malloc(strlen(pb)+1);
          if(str_tab[nn])
            strcpy(str_tab[nn],pb);
        }
      else
        SetDlgItemText(hWnd,nn,pb);
    }
  fclose(in);
}

char* LoadText(int id)
{
  static char buf[80];

  if((id>=IDS_BEGIN)&&(id<=IDS_END)&&(str_tab[id-IDS_BEGIN]))
    return str_tab[id-IDS_BEGIN];

  LoadString(hInst,id,buf,sizeof(buf));
  return buf;
}

void PrintError(HWND hWnd,DWORD msg)
{
  MessageBox(hWnd,LoadText(msg),NULL,MB_OK | MB_ICONERROR);
}

char* flipAndCodeBytes(const char* str,int pos,int flip,char* buf)
{
   int i;
   int j=0;
   int k=0;

   buf [0]='\0';
   if(pos <=0)
     return buf;

   if(!j)
   {
     char p=0;

     j=1;
     k=0;
     buf[k]=0;
     for(i=pos;j&&str[i]!='\0';++i)
     {
       char c=tolower(str[i]);

      if(isspace(c))
        c='0';

      ++p;
      buf[k] <<=4;

      if(c>='0'&&c<='9')
         buf[k]|=(unsigned char)(c - '0');
      else if(c>='a'&&c<='f')
         buf[k]|=(unsigned char)(c - 'a' + 10);
      else
      {
        j=0;
        break;
      }

      if(p==2)
      {
         if(buf[k] !='\0' && !isprint(buf[k]))
         {
           j=0;
           break;
          }
         ++k;
         p=0;
         buf[k]=0;
      }

      }
   }

   if(!j)
   {
     j=1;
     k=0;
     for(i=pos; j && str[i] !='\0'; ++i)
     {
       char c=str[i];

       if(!isprint(c))
         {
           j=0;
           break;
        }

       buf[k++]=c;
      }
   }

   if(!j)
     k=0;

   buf[k]='\0';

   if(flip)
     for(j=0; j < k; j +=2)
     {
       char t=buf[j];
       buf[j]=buf[j + 1];
       buf[j + 1]=t;
     }

   i=j=-1;
   for(k=0; buf[k] !='\0'; ++k)
   {
     if(!isspace(buf[k]))
     {
       if(i < 0)
       i=k;
       j=k;
      }
   }

  if((i>=0)&&(j>=0))
  {
    for(k=i;(k<=j)&&(buf[k]!='\0');++k)
      buf[k-i]=buf[k];
    buf[k-i]='\0';
  }
  return buf;
}

typedef struct _STORAGE_DEVICE_NUMBER {
  DEVICE_TYPE DeviceType;
  ULONG       DeviceNumber;
  ULONG       PartitionNumber;
} STORAGE_DEVICE_NUMBER, *PSTORAGE_DEVICE_NUMBER;

void RefreshDisk(HWND hWnd)
{
  int i;
  int idx=-1;
  char vl[24];
  STORAGE_DEVICE_NUMBER d1;
  DWORD nr;
  HANDLE hd;
  char buf[6];

  for(i=0;i<MAX_DISKS;i++)
    strcpy(hdl[i],"");

  for(i=(byte)'C';i<=(byte)'Z';i++)
  {
    sprintf(vl,"\\\\.\\%c:",(char)i);
    hd=CreateFile(vl,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,OPEN_EXISTING,0,0);
    if(hd==INVALID_HANDLE_VALUE)
      continue;

    if(!DeviceIoControl(hd,IOCTL_STORAGE_GET_DEVICE_NUMBER,NULL,0,&d1,sizeof(d1),&nr,0))
    {
      CloseHandle(hd);
      continue;
    }
    if(d1.DeviceType==FILE_DEVICE_DISK)
    {
      if(!strcmp(hdl[d1.DeviceNumber],""))
        sprintf(buf,"%c",(char)i);
      else
        sprintf(buf,", %c",(char)i);
      strcat(hdl[d1.DeviceNumber],buf);
    }
    CloseHandle(hd);
  }

  char dn[24]="\\\\.\\PhysicalDrive/",nn[256];

  SendDlgItemMessage(hWnd,IDC_DISKS,CB_RESETCONTENT,0,0);
  for(i=0;i<MAX_DISKS;i++)
    {
      HANDLE hd;
      GET_LENGTH_INFORMATION lpOutBuffer;
      STORAGE_PROPERTY_QUERY storagePropertyQuery;
      PSTORAGE_DEVICE_DESCRIPTOR descrip;
      char prinfo[20001]="";
      DWORD lpBytesReturned; 
      DOUBLE sz;
      char mu[3]="GB";
      char buffer[10000];
      char vendorID[10000];
      char modelNumber[10000];

      if(i < 10)
               ++dn[17];
          else if(i > 10)
               ++dn[18];
          else
               *(unsigned short *)&dn[17]=0x3031;//or use dn[17]='1',dn[18]='0',dn[19]=0;

      hd=CreateFile(dn,GENERIC_READ | GENERIC_WRITE,FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,OPEN_EXISTING, 0, NULL);
      if(hd==INVALID_HANDLE_VALUE)
        continue;

      DWORD outsize;
      STORAGE_PROPERTY_QUERY spq;
      STORAGE_DEVICE_DESCRIPTOR sdd = {0};
      spq.PropertyId = StorageDeviceProperty;
      spq.QueryType = PropertyStandardQuery;
      memset(&sdd, 0, sizeof(sdd));
      char bustype[10]="";

      if(DeviceIoControl(hd,IOCTL_STORAGE_QUERY_PROPERTY,&spq,sizeof(STORAGE_PROPERTY_QUERY),&sdd,sizeof(STORAGE_DEVICE_DESCRIPTOR),&outsize,NULL))
      {
        switch (sdd.BusType)
        {
          case 0:
            strcpy(bustype,"UNKNOWN  ");
            break;
          case 1:
            strcpy(bustype,"SCSI  ");
            break;
          case 2:
            strcpy(bustype,"ATAPI  ");
            break;
          case 3:
            strcpy(bustype,"ATA  ");
            break;
          case 4:
            strcpy(bustype,"1394  ");
            break;
          case 5:
            strcpy(bustype,"SSA  ");
            break;
          case 6:
            strcpy(bustype,"FIBRE  ");
            break;
          case 7:
            strcpy(bustype,"USB  ");
            break;
          case 8:
            strcpy(bustype,"RAID  ");
            break;
          case 9:
            strcpy(bustype,"iSCSI  ");
            break;
          case 10:
            strcpy(bustype,"SAS  ");
            break;
          case 11:
            strcpy(bustype,"SATA  ");
            break;
          case 12:
            strcpy(bustype,"SD  ");
            break;
          case 13:
            strcpy(bustype,"MMC  ");
            break;
          default:
            strcpy(bustype,"");
        }
      }

      sz=0;
      if(DeviceIoControl(hd,IOCTL_DISK_GET_LENGTH_INFO,NULL,0,&lpOutBuffer,sizeof(GET_LENGTH_INFORMATION),
        (LPDWORD)&lpBytesReturned,NULL))
        {
          sz=1.0/1073741824*lpOutBuffer.Length.QuadPart;
          if(sz<1)
          {
            sz*=1024;
            strcpy(mu,"MB");
          }
          else
            if(sz>=1024)
            {
              sz/=1024;
              strcpy(mu,"TB");
            }
        }

        ZeroMemory(&storagePropertyQuery,sizeof(STORAGE_PROPERTY_QUERY));
        storagePropertyQuery.PropertyId=StorageDeviceProperty;
        storagePropertyQuery.QueryType=PropertyStandardQuery;
        FillMemory(buffer, sizeof(buffer), 0);
        lpBytesReturned=0;
        if(DeviceIoControl(hd, IOCTL_STORAGE_QUERY_PROPERTY, &storagePropertyQuery, sizeof(STORAGE_PROPERTY_QUERY),
            buffer, sizeof(buffer),(LPDWORD)&lpBytesReturned, NULL))
        {
          descrip=(PSTORAGE_DEVICE_DESCRIPTOR)&buffer;
          flipAndCodeBytes(buffer,descrip->VendorIdOffset,0,vendorID);
          strcpy(prinfo,vendorID);
          strcat(prinfo," ");
          flipAndCodeBytes(buffer,descrip->ProductIdOffset,0,modelNumber);
          strcat(prinfo,modelNumber);
        }

        if(strlen(prinfo)<=1)
        {
          if(sz>=100)
            sprintf(nn,"hd%d: %sUnknown, %.0f %s [%s]]",i,bustype,sz,mu,hdl[i]);
          else
            if(sz>=10)
              sprintf(nn,"hd%d: %sUnknown, %.1f %s [%s]",i,bustype,sz,mu,hdl[i]);
            else
              if(sz>=1)
                sprintf(nn,"hd%d: %sUnknown, %.2f %s [%s]",i,bustype,sz,mu,hdl[i]);
              else
                if(sz>=0.1)
                  sprintf(nn,"hd%d: %sUnknown, %.3f %s [%s]",i,bustype,sz,mu,hdl[i]);
                else
                  sprintf(nn,"hd%d: %sUnknown",i,bustype);
         }
         else
        {
          if(sz>=100)
            sprintf(nn,"hd%d: %s%s, %.0f %s [%s]",i,bustype,prinfo,sz,mu,hdl[i]);
          else
            if(sz>=10)
              sprintf(nn,"hd%d: %s%s, %.1f %s [%s]",i,bustype,prinfo,sz,mu,hdl[i]);
            else
              if(sz>=1)
                sprintf(nn,"hd%d: %s%s, %.2f %s [%s]",i,bustype,prinfo,sz,mu,hdl[i]);
              else
                if(sz>=0.1)
                  sprintf(nn,"hd%d: %s%s, %.3f %s [%s]",i,bustype,prinfo,sz,mu,hdl[i]);
                else
                  sprintf(nn,"hd%d: %s%s",i,bustype,prinfo);
         }

      SendDlgItemMessage(hWnd,IDC_DISKS,CB_ADDSTRING,0,(LPARAM)&nn);
      idx++;
      hdn[idx]=i;
      CloseHandle(hd);
    }
}

int GetFileName(HWND hWnd,char* fn,int len,BOOLEAN addqts)
{
  if(IsDlgButtonChecked(hWnd,IDC_ISDISK)==BST_CHECKED)
    {
      char buf[512];
      int idx;

      if(GetDlgItemText(hWnd,IDC_DISKS,buf,len)==0)
        {
          PrintError(hWnd,IDS_NO_DISKNAME);
          return 1;
        }
      idx=SendDlgItemMessage(hWnd,IDC_DISKS,CB_GETCURSEL,0,0);
      if(idx>-1)
      {
        sprintf(fn,"(hd%d)",hdn[idx]);
        len=strlen(fn);
      }
    }
  else if(IsDlgButtonChecked(hWnd,IDC_ISFILE)==BST_CHECKED)
    {
      if(GetDlgItemText(hWnd,IDC_FILENAME,fn,len)==0)
        {
          PrintError(hWnd,IDS_NO_FILENAME);
          return 1;
        }
      if(addqts)
        {
          char buf[MAX_PATH+2];

          sprintf(buf,"\"%s\"",fn);
          strcpy(fn,&buf[0]);
          len+=2;
        }
    }
  else
    {
      PrintError(hWnd,IDS_NO_DEVICETYPE);
      return 1;
    }
  return 0;
}

void RefreshPart(HWND hWnd,BOOLEAN silent)
{
  char fn[MAX_PATH];
  char buf[512*4];
  char temp[128];
  int fs,len;
  xd_t *xd;
  DOUBLE sz;
  char mu[3]="GB";

  SendDlgItemMessage(hWnd,IDC_PARTS,CB_RESETCONTENT,0,0);

  if(GetFileName(hWnd,fn,sizeof(fn),FALSE))
  {
    if(IsDlgButtonChecked(hWnd,IDC_ISDISK)==BST_CHECKED)
      SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_DISKS),TRUE);
    else
      if(IsDlgButtonChecked(hWnd,IDC_ISFILE)==BST_CHECKED)
        SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_FILENAME),TRUE);
    return;
  }

  xd=xd_open(fn,0);
  if(xd==NULL)
    {
      if(!silent)
      {
        PrintError(hWnd,IDS_INVALID_FILE);
        if(IsDlgButtonChecked(hWnd,IDC_ISDISK)==BST_CHECKED)
          SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_DISKS),TRUE);
        else
          if(IsDlgButtonChecked(hWnd,IDC_ISFILE)==BST_CHECKED)
            SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_FILENAME),TRUE);
      }
      return;
    }
  if(xd_read(xd,buf,sizeof(buf)>>9))
    {
      if(!silent)
      {
        PrintError(hWnd,IDS_INVALID_FILE);
        if(IsDlgButtonChecked(hWnd,IDC_ISDISK)==BST_CHECKED)
          SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_DISKS),TRUE);
        else
          if(IsDlgButtonChecked(hWnd,IDC_ISFILE)==BST_CHECKED)
            SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_FILENAME),TRUE);
      }
      xd_close(xd);
      return;
    }
  GetDlgItemText(hWnd,IDC_FILESYSTEM,temp,sizeof(temp));
  if(!strcmp(temp,"MBR"))
    fs=FST_MBR;
  else
  if(!strcmp(temp,"NTFS"))
    fs=FST_NTFS;
  else
  if(!strcmp(temp,"FAT32"))
    fs=FST_FAT32;
  else
  if(!strcmp(temp,"FAT12/FAT16"))
    fs=FST_FAT16;
  else
  if(!strcmp(temp,"EXT2/EXT3"))
    fs=FST_EXT2;
  else
    fs=get_fstype(buf);

  if(fs==FST_MBR2)
    {
      if(!silent)
        PrintError(hWnd, IDS_INVALID_MBR);
      fs=FST_MBR;
    }

  CheckDlgButton(hWnd,IDC_IS_FLOPPY,(fs!=FST_MBR));

  strcpy(buf,LoadText(IDS_WHOLE_DISK));
  len=strlen(buf);
  sprintf(&buf[len]," (%s)",fst2str(fs));
  SendDlgItemMessage(hWnd,IDC_PARTS,CB_ADDSTRING,0,(LPARAM)buf);

  if(fs==FST_MBR)
    {
      xde_t xe;

      xe.cur=xe.nxt=0xFF;
      while(!xd_enum(xd,&xe))
        {
          sz=1.0/2097152*xe.len;
          if(sz<1)
          {
            sz*=1024;
            strcpy(mu,"MB");
          }
          else
            if(sz>=1024)
            {
              sz/=1024;
              strcpy(mu,"TB");
            }
          if(sz>=100)
            sprintf(buf,"p%d: %02X(%s) [%.0f %s]",xe.cur,xe.dfs,dfs2str(xe.dfs),sz,mu);
          else
            if(sz>=10)
              sprintf(buf,"p%d: %02X(%s) [%.1f %s]",xe.cur,xe.dfs,dfs2str(xe.dfs),sz,mu);
            else
              if(sz>=1)
                sprintf(buf,"p%d: %02X(%s) [%.2f %s]",xe.cur,xe.dfs,dfs2str(xe.dfs),sz,mu);
              else
                if(sz>=0.1)
                  sprintf(buf,"p%d: %02X(%s) [%.3f %s]",xe.cur,xe.dfs,dfs2str(xe.dfs),sz,mu);
                else
                  if(sz>=0.01)
                    sprintf(buf,"p%d: %02X(%s) [%.4f %s]",xe.cur,xe.dfs,dfs2str(xe.dfs),sz,mu);
                  else
                    sprintf(buf,"p%d: %02X(%s)",xe.cur,xe.dfs,dfs2str(xe.dfs));
          SendDlgItemMessage(hWnd,IDC_PARTS,CB_ADDSTRING,0,(LPARAM)buf);
        }
    }
  xd_close(xd);
  SendDlgItemMessage(hWnd,IDC_PARTS,CB_SETCURSEL,(WPARAM)0,0);
}

void Install(HWND hWnd,BOOLEAN test)
{
  char fn[512],*pb,sfn[512];
  char temp[16];
  int len;
  int i;
  char osdrv[7]="";

  if(!test)
  {
    DWORD nr;
    char vl[24];
    STORAGE_DEVICE_NUMBER d1;
    HANDLE hd;
    char osdrvl[3]="\1\1";

    if(GetEnvironmentVariable("SystemDrive",osdrvl,sizeof(osdrvl)))
      for(i=(byte)'C';i<=(byte)'Z';i++)
        if((char)i==osdrvl[0])
        {
          sprintf(vl,"\\\\.\\%c:",(char)i);
          hd=CreateFile(vl,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,OPEN_EXISTING,0,0);
          if(hd==INVALID_HANDLE_VALUE)
            break;
 
          if(!DeviceIoControl(hd,IOCTL_STORAGE_GET_DEVICE_NUMBER,NULL,0,&d1,sizeof(d1),&nr,0))
          {
            CloseHandle(hd);
            break;
          }
          if(d1.DeviceType==FILE_DEVICE_DISK)
            sprintf(osdrv,"(hd%d)",(int)d1.DeviceNumber);
          CloseHandle(hd);
          break;
       }
  }

  GetModuleFileName(hInst,fn,sizeof(fn));
  strcpy(sfn,fn);
  pb=strrchr(sfn,'\\');
  if(!pb)
    pb=sfn;
  else
    pb++;
  strcpy(pb,"grubinst.exe\0");
  pb=strrchr(fn,'\\');
  if(!pb)
    pb=fn;
  else
    pb++;
  strcpy(pb,"grubinst.exe --pause ");
  len=strlen(fn);

  if(test)
    {
      strcpy(&fn[len],"--read-only ");
      len+=strlen(&fn[len]);
    }
  GetDlgItemText(hWnd,IDC_FILESYSTEM,temp,sizeof(temp));
  if(!strcmp(temp,"MBR"))
    {
      strcpy(&fn[len],"--file-system-type=MBR ");
      len+=strlen(&fn[len]);
    }
  else
  if(!strcmp(temp,"NTFS"))
    {
      strcpy(&fn[len],"--file-system-type=NTFS ");
      len+=strlen(&fn[len]);
    }
  else
  if(!strcmp(temp,"FAT32"))
    {
      strcpy(&fn[len],"--file-system-type=FAT32 ");
      len+=strlen(&fn[len]);
    }
  else
  if(!strcmp(temp,"FAT12/FAT16"))
    {
      strcpy(&fn[len],"--file-system-type=FAT16 ");
      len+=strlen(&fn[len]);
    }
  else
  if(!strcmp(temp,"EXT2/EXT3"))
    {
      strcpy(&fn[len],"--file-system-type=EXT2 ");
      len+=strlen(&fn[len]);
    }
  if(IsDlgButtonChecked(hWnd,IDC_OUTPUT)==BST_CHECKED)
    {
      strcpy(&fn[len],"--output ");
      len+=strlen(&fn[len]);
    } 
  if(IsDlgButtonChecked(hWnd,IDC_VERBOSE)==BST_CHECKED)
    {
      strcpy(&fn[len],"--verbose ");
      len+=strlen(&fn[len]);
    }
  if(IsDlgButtonChecked(hWnd,IDC_NO_BACKUP_MBR)==BST_CHECKED)
    {
      strcpy(&fn[len],"--no-backup-mbr ");
      len+=strlen(&fn[len]);
    }
  if(IsDlgButtonChecked(hWnd,IDC_DISABLE_FLOPPY)==BST_CHECKED)
    {
      strcpy(&fn[len],"--mbr-disable-floppy ");
      len+=strlen(&fn[len]);
    }
  if(IsDlgButtonChecked(hWnd,IDC_DISABLE_OSBR)==BST_CHECKED)
    {
      strcpy(&fn[len],"--mbr-disable-osbr ");
      len+=strlen(&fn[len]);
    }
  if(IsDlgButtonChecked(hWnd,IDC_PREVMBR_FIRST)==BST_CHECKED)
    {
      strcpy(&fn[len],"--boot-prevmbr-first ");
      len+=strlen(&fn[len]);
    }
  if(IsDlgButtonChecked(hWnd,IDC_GRUB2)==BST_CHECKED)
    {
      strcpy(&fn[len],"--grub2 ");
      len+=strlen(&fn[len]);
    }
  if(IsDlgButtonChecked(hWnd,IDC_IS_FLOPPY)==BST_CHECKED)
    {
      strcpy(&fn[len],"--floppy ");
      len+=strlen(&fn[len]);
    }
  if(IsDlgButtonChecked(hWnd,IDC_SKIP)==BST_CHECKED)
    {
      strcpy(&fn[len],"--skip-mbr-test ");
      len+=strlen(&fn[len]);
    }
  if(IsDlgButtonChecked(hWnd,IDC_SILENT)==BST_CHECKED)
    {
      strcpy(&fn[len],"--silent-boot ");
      len+=strlen(&fn[len]);
    }
  if(GetDlgItemText(hWnd,IDC_TIMEOUT,temp,sizeof(temp))!=0)
    {
      sprintf(&fn[len],"--time-out=%s ",temp);
      len+=strlen(&fn[len]);
    }
  else
    {
      sprintf(&fn[len],"--time-out=0 ");
      len+=strlen(&fn[len]);
    }
  if(GetDlgItemText(hWnd,IDC_HOTKEY,temp,sizeof(temp))!=0)
    {
      sprintf(&fn[len],"--hot-key=%s ",temp);
      len+=strlen(&fn[len]);
    }
  if(GetDlgItemText(hWnd,IDC_LOADSEG,temp,sizeof(temp))!=0)
    {
      sprintf(&fn[len],"--load-seg=%s ",temp);
      len+=strlen(&fn[len]);
    }
  if(GetDlgItemText(hWnd,IDC_BOOTFILE,temp,sizeof(temp))!=0)
    {
      sprintf(&fn[len],"--boot-file=%s ",temp);
      len+=strlen(&fn[len]);
    }
  if(IsDlgButtonChecked(hWnd,IDC_RESTOREMBR)==BST_CHECKED)
    {
      strcpy(&fn[len],"\"--restore=");
      len+=strlen(&fn[len]);
      if(GetDlgItemText(hWnd,IDC_SAVEFILE,&fn[len],sizeof(fn)-len)==0)
        {
          PrintError(hWnd,IDS_NO_RESTOREFILE);
          SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_SAVEFILE),TRUE);
          return;
        }
      len+=strlen(&fn[len]);
      fn[len++]='"';
      fn[len++]=' ';
      fn[len]=0;
    }
  else if(IsDlgButtonChecked(hWnd,IDC_RESTORE_PREVMBR)==BST_CHECKED)
    {
      strcpy(&fn[len],"--restore-prevmbr ");
      len+=strlen(&fn[len]);
    }
  else
    {
      int slen;

      if(IsDlgButtonChecked(hWnd,IDC_SAVEMBR)==BST_CHECKED)
      {
        strcpy(&fn[len],"\"--save=");
        slen=strlen(&fn[len]);
        if(GetDlgItemText(hWnd,IDC_SAVEFILE,&fn[len+slen],sizeof(fn)-len-slen)!=0)
        {
          len+=strlen(&fn[len]);
          fn[len++]='"';
          fn[len++]=' ';
          fn[len]=0;
        }
        else
        {
          PrintError(hWnd,IDS_NO_SAVEFILE);
          SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_SAVEFILE),TRUE);
          return;
        }         
      }
    }
  if(GetDlgItemText(hWnd,IDC_PARTS,temp,sizeof(temp))!=0)
    if((temp[1]>='0')&&(temp[1]<='9'))
      {
        int n;

        n=atoi(&temp[1]);
        sprintf(&fn[len],"--install-partition=%d ",n);
        len+=strlen(&fn[len]);
      }
  if(GetDlgItemText(hWnd,IDC_EXTRA,&fn[len],sizeof(fn)-len)!=0)
    {
      len+=strlen(&fn[len]);
      fn[len++]=' ';
      fn[len]=0;
    }
  if(IsDlgButtonChecked(hWnd,IDC_OUTPUT)==BST_UNCHECKED)
  {
    if(GetFileName(hWnd,&fn[len],sizeof(fn)-len,TRUE))
    {
      if(IsDlgButtonChecked(hWnd,IDC_ISDISK)==BST_CHECKED)
        SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_DISKS),TRUE);
      else
        if(IsDlgButtonChecked(hWnd,IDC_ISFILE)==BST_CHECKED)
          SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_FILENAME),TRUE);
      return;
    }
    if((!test)&&(!strcmp(&fn[len],osdrv)))
      if(MessageBox(hWnd,LoadText(IDS_OSDRIVE),"Warning",MB_YESNO | MB_ICONWARNING | MB_DEFBUTTON2)!=IDYES)
        return;
  }
  else
  {
    strcpy(&fn[len],"\"");
    len+=strlen(&fn[len]);
    if(GetDlgItemText(hWnd,IDC_SAVEFILE,&fn[len],sizeof(fn)-len)==0)
    {
      PrintError(hWnd,IDS_NO_SAVEFILE);
      SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_SAVEFILE),TRUE);
      return;
    }
    len+=strlen(&fn[len]);
    fn[len++]='"';
    fn[len]=0;
  }

  FILE * file;
  file=fopen(sfn,"r");
  if(file)
  {
    fclose(file);
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    ZeroMemory(&si,sizeof(si));
    si.cb=sizeof(si);
    ZeroMemory(&pi, sizeof(pi));
    if(test)
    {
      char cmlwp[1024];
 
      strcpy(cmlwp,"cmd.exe /c @echo Starting \"grubinst.exe");
      strcpy(&cmlwp[39],&fn[strlen(sfn)]);
      strcpy(&cmlwp[39+strlen(fn)-strlen(sfn)], "\"... && echo. && \"");
      strcpy(&cmlwp[57+strlen(fn)-strlen(sfn)],sfn);
      strcpy(&cmlwp[57+strlen(fn)],"\"");
      strcpy(&cmlwp[58+strlen(fn)],&fn[strlen(sfn)]);
      if(!CreateProcess(NULL,cmlwp,NULL,NULL,FALSE,CREATE_NEW_CONSOLE,NULL,NULL,&si,&pi))
      {
        char errmsg[512];
        sprintf(errmsg,"%s\r\nError code: %d",LoadText(IDS_EXEC_ERROR),(int)GetLastError());
        MessageBox(hWnd,errmsg,NULL,MB_OK | MB_ICONERROR);
        return;
      }
    }
    else
      if(!CreateProcess(NULL,fn,NULL,NULL,FALSE,CREATE_NEW_CONSOLE,NULL,NULL,&si,&pi))
      {
        char errmsg[512];
        sprintf(errmsg,"%s\r\nError code: %d",LoadText(IDS_EXEC_ERROR),(int)GetLastError());
        MessageBox(hWnd,errmsg,NULL,MB_OK | MB_ICONERROR);
        return;
      }
    WaitForSingleObject(pi.hProcess,INFINITE);
    CloseHandle(pi.hProcess);
    CloseHandle(pi.hThread);
  }
  else
    MessageBox(hWnd,LoadText(IDS_NO_GRUBINST),NULL,MB_OK | MB_ICONERROR);
}

BOOL CALLBACK DialogOutputProc(HWND hOWnd,UINT wMsg,WPARAM wParam,LPARAM lParam)
{

  switch(wMsg){
  case WM_CLOSE:
    {
      EndDialog(hOWnd,0);
      break;
    }
  case WM_CONTEXTMENU:
    {
      if((HWND)wParam!=GetDlgItem(hOWnd,IDC_RICHEDIT))
        break;
      HMENU hMenu;
      CHARRANGE cr;
      int len;
      GETTEXTLENGTHEX tls;
 
      hMenu = CreatePopupMenu();
      SendMessage((HWND)wParam, EM_EXGETSEL, 0,(LPARAM)&cr);
      if(cr.cpMax>cr.cpMin)
        InsertMenuW(hMenu, 0, MF_STRING | MF_BYPOSITION, IDC_COPY, L"Copy");
      else
        InsertMenuW(hMenu, 0, MF_STRING | MF_BYPOSITION | MF_GRAYED, IDC_COPY, L"Copy");
      tls.flags=GTL_NUMCHARS;
      tls.codepage=CP_ACP;
      len=SendMessage((HWND)wParam, EM_GETTEXTLENGTHEX,(WPARAM)&tls,(LPARAM)0);
      if((cr.cpMin==0)&&(cr.cpMax==(len+1)))
        InsertMenuW(hMenu, 1, MF_STRING | MF_BYPOSITION | MF_GRAYED, IDC_SELALL, L"Select all");
      else
        InsertMenuW(hMenu, 1, MF_STRING | MF_BYPOSITION, IDC_SELALL, L"Select all");
      TrackPopupMenu(hMenu, TPM_LEFTALIGN | TPM_RIGHTBUTTON, GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam), 0, hOWnd, NULL); 
      DestroyMenu(hMenu); 
      break;
    }
  case WM_INITDIALOG:
    {
      HICON hIcon;
      char title[30];
      char buf[30];
      HWND hWnd;

      hWnd=GetParent(hOWnd);
      if(lParam==0)
        sprintf(title,"Output Test");
      else
      {
        GetDlgItemText(hWnd,IDC_INSTALL,buf,sizeof(buf));
        sprintf(title,"Output %s",buf);
      }
      SendMessage(hOWnd,WM_SETTEXT,0,(LPARAM)title);
      hIcon=LoadIcon(hInst,MAKEINTRESOURCE(ICO_MAIN));
      SendMessage(hOWnd,WM_SETICON,ICON_BIG,(LPARAM)hIcon);

      char fn[512],*pb,sfn[512];
      char temp[16];
      int len;
      int i;
      char osdrv[7]="";

      if(lParam==1)
      {
        DWORD nr;
        char vl[24];
        STORAGE_DEVICE_NUMBER d1;
        HANDLE hd;
        char osdrvl[3]="\1\1";

        if(GetEnvironmentVariable("SystemDrive",osdrvl,sizeof(osdrvl)))
          for(i=(byte)'C';i<=(byte)'Z';i++)
            if((char)i==osdrvl[0])
            {
              sprintf(vl,"\\\\.\\%c:",(char)i);
              hd=CreateFile(vl,GENERIC_READ,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,OPEN_EXISTING,0,0);
              if(hd==INVALID_HANDLE_VALUE)
                break;
    
              if(!DeviceIoControl(hd,IOCTL_STORAGE_GET_DEVICE_NUMBER,NULL,0,&d1,sizeof(d1),&nr,0))
              {
                CloseHandle(hd);
                break;
              }
              if(d1.DeviceType==FILE_DEVICE_DISK)
                sprintf(osdrv,"(hd%d)",(int)d1.DeviceNumber);
              CloseHandle(hd);
              break;
            }
      }
 
      GetModuleFileName(hInst,fn,sizeof(fn));
      strcpy(sfn,fn);
      pb=strrchr(sfn,'\\');
      if(!pb)
        pb=sfn;
      else
        pb++;
      strcpy(pb,"grubinst.exe\0");
      pb=strrchr(fn,'\\');
      if(!pb)
        pb=fn;
      else
        pb++;
      strcpy(pb,"grubinst.exe ");
      len=strlen(fn);
 
      if(lParam==0)
        {
          strcpy(&fn[len],"--read-only ");
          len+=strlen(&fn[len]);
        }
      GetDlgItemText(hWnd,IDC_FILESYSTEM,temp,sizeof(temp));
      if(!strcmp(temp,"MBR"))
        {
          strcpy(&fn[len],"--file-system-type=MBR ");
          len+=strlen(&fn[len]);
        }
      else
      if(!strcmp(temp,"NTFS"))
        {
          strcpy(&fn[len],"--file-system-type=NTFS ");
          len+=strlen(&fn[len]);
        }
      else
      if(!strcmp(temp,"FAT32"))
        {
          strcpy(&fn[len],"--file-system-type=FAT32 ");
          len+=strlen(&fn[len]);
        }
      else
      if(!strcmp(temp,"FAT12/FAT16"))
        {
          strcpy(&fn[len],"--file-system-type=FAT16 ");
          len+=strlen(&fn[len]);
        }
      else
      if(!strcmp(temp,"EXT2/EXT3"))
        {
          strcpy(&fn[len],"--file-system-type=EXT2 ");
          len+=strlen(&fn[len]);
        }
      if(IsDlgButtonChecked(hWnd,IDC_OUTPUT)==BST_CHECKED)
        {
          strcpy(&fn[len],"--output ");
          len+=strlen(&fn[len]);
        }    
      if(IsDlgButtonChecked(hWnd,IDC_VERBOSE)==BST_CHECKED)
        {
          strcpy(&fn[len],"--verbose ");
          len+=strlen(&fn[len]);
        }
      if(IsDlgButtonChecked(hWnd,IDC_NO_BACKUP_MBR)==BST_CHECKED)
        {
          strcpy(&fn[len],"--no-backup-mbr ");
          len+=strlen(&fn[len]);
        }
      if(IsDlgButtonChecked(hWnd,IDC_DISABLE_FLOPPY)==BST_CHECKED)
        {
          strcpy(&fn[len],"--mbr-disable-floppy ");
          len+=strlen(&fn[len]);
        }
      if(IsDlgButtonChecked(hWnd,IDC_DISABLE_OSBR)==BST_CHECKED)
        {
          strcpy(&fn[len],"--mbr-disable-osbr ");
          len+=strlen(&fn[len]);
        }
      if(IsDlgButtonChecked(hWnd,IDC_PREVMBR_FIRST)==BST_CHECKED)
        {
          strcpy(&fn[len],"--boot-prevmbr-first ");
          len+=strlen(&fn[len]);
        }
      if(IsDlgButtonChecked(hWnd,IDC_GRUB2)==BST_CHECKED)
        {
          strcpy(&fn[len],"--grub2 ");
          len+=strlen(&fn[len]);
        }
      if(IsDlgButtonChecked(hWnd,IDC_IS_FLOPPY)==BST_CHECKED)
        {
          strcpy(&fn[len],"--floppy ");
          len+=strlen(&fn[len]);
        }
      if(IsDlgButtonChecked(hWnd,IDC_SKIP)==BST_CHECKED)
        {
          strcpy(&fn[len],"--skip-mbr-test ");
          len+=strlen(&fn[len]);
        }
      if(IsDlgButtonChecked(hWnd,IDC_SILENT)==BST_CHECKED)
        {
          strcpy(&fn[len],"--silent-boot ");
          len+=strlen(&fn[len]);
        }
      if(GetDlgItemText(hWnd,IDC_TIMEOUT,temp,sizeof(temp))!=0)
        {
          sprintf(&fn[len],"--time-out=%s ",temp);
          len+=strlen(&fn[len]);
        }
      else
        {
          sprintf(&fn[len],"--time-out=0 ");
          len+=strlen(&fn[len]);
        }
      if(GetDlgItemText(hWnd,IDC_HOTKEY,temp,sizeof(temp))!=0)
        {
          sprintf(&fn[len],"--hot-key=%s ",temp);
          len+=strlen(&fn[len]);
        }
      if(GetDlgItemText(hWnd,IDC_LOADSEG,temp,sizeof(temp))!=0)
        {
          sprintf(&fn[len],"--load-seg=%s ",temp);
          len+=strlen(&fn[len]);
        }
      if(GetDlgItemText(hWnd,IDC_BOOTFILE,temp,sizeof(temp))!=0)
        {
          sprintf(&fn[len],"--boot-file=%s ",temp);
          len+=strlen(&fn[len]);
        }
      if(IsDlgButtonChecked(hWnd,IDC_RESTOREMBR)==BST_CHECKED)
        {
          strcpy(&fn[len],"\"--restore=");
          len+=strlen(&fn[len]);
          if(GetDlgItemText(hWnd,IDC_SAVEFILE,&fn[len],sizeof(fn)-len)==0)
            {
              PrintError(hWnd,IDS_NO_RESTOREFILE);
              EndDialog(hOWnd,0);
              SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_SAVEFILE),TRUE);
              break;
            }
          len+=strlen(&fn[len]);
          fn[len++]='"';
          fn[len++]=' ';
          fn[len]=0;
        }
      else if(IsDlgButtonChecked(hWnd,IDC_RESTORE_PREVMBR)==BST_CHECKED)
        {
          strcpy(&fn[len],"--restore-prevmbr ");
          len+=strlen(&fn[len]);
        }
      else
        {
          int slen;

          if(IsDlgButtonChecked(hWnd,IDC_SAVEMBR)==BST_CHECKED)    
          {
            strcpy(&fn[len],"\"--save=");
            slen=strlen(&fn[len]);
            if(GetDlgItemText(hWnd,IDC_SAVEFILE,&fn[len+slen],sizeof(fn)-len-slen)!=0)
            {
              len+=strlen(&fn[len]);
              fn[len++]='"';
              fn[len++]=' ';
              fn[len]=0;
            }
            else
            {
              PrintError(hWnd,IDS_NO_SAVEFILE);
              EndDialog(hOWnd,0);
              SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_SAVEFILE),TRUE);
              break;
            }
          }
        }
      if(GetDlgItemText(hWnd,IDC_PARTS,temp,sizeof(temp))!=0)
        if((temp[1]>='0')&&(temp[1]<='9'))
          {
            int n;
    
            n=atoi(&temp[1]);
            sprintf(&fn[len],"--install-partition=%d ",n);
            len+=strlen(&fn[len]);
          }
      if(GetDlgItemText(hWnd,IDC_EXTRA,&fn[len],sizeof(fn)-len)!=0)
        {
          len+=strlen(&fn[len]);
          fn[len++]=' ';
          fn[len]=0;
        }
      if(IsDlgButtonChecked(hWnd,IDC_OUTPUT)==BST_UNCHECKED)
      {
        if(GetFileName(hWnd,&fn[len],sizeof(fn)-len,TRUE))
          {
            EndDialog(hOWnd,0);
            if(IsDlgButtonChecked(hWnd,IDC_ISDISK)==BST_CHECKED)
              SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_DISKS),TRUE);
            else
               if(IsDlgButtonChecked(hWnd,IDC_ISFILE)==BST_CHECKED)
                 SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_FILENAME),TRUE);
            break;
          }
        if((lParam==1)&&(!strcmp(&fn[len],osdrv)))
          if(MessageBox(hWnd,LoadText(IDS_OSDRIVE),"Warning",MB_YESNO | MB_ICONWARNING | MB_DEFBUTTON2)!=IDYES)
          {
            EndDialog(hOWnd,0);
            break;
          }
        }
        else
          {
            strcpy(&fn[len],"\"");
            len+=strlen(&fn[len]);
            if(GetDlgItemText(hWnd,IDC_SAVEFILE,&fn[len],sizeof(fn)-len)==0)
              {
                PrintError(hWnd,IDS_NO_SAVEFILE);
                EndDialog(hOWnd,0);
                SendMessage(hWnd,WM_NEXTDLGCTL,(WPARAM)GetDlgItem(hWnd,IDC_SAVEFILE),TRUE);
                break;
              }
            len+=strlen(&fn[len]);
            fn[len++]='"';
            fn[len]=0;
          }
    
      FILE * file;
      file=fopen(sfn,"r");
      if(file)
      {
        fclose(file);

        char outtxt[10240]="";
        char buf[1024];
        SECURITY_ATTRIBUTES sa;
        SECURITY_DESCRIPTOR sd;
        STARTUPINFO si;
        PROCESS_INFORMATION pi;
        HANDLE newstdout,read_stdout;
    
        InitializeSecurityDescriptor(&sd,SECURITY_DESCRIPTOR_REVISION);
        SetSecurityDescriptorDacl(&sd,TRUE,NULL,FALSE);
        sa.lpSecurityDescriptor=&sd;
        sa.nLength=sizeof(SECURITY_ATTRIBUTES);
        sa.bInheritHandle=TRUE;
        CreatePipe(&read_stdout,&newstdout,&sa,0);
        GetStartupInfo(&si);
        si.dwFlags=STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
        si.wShowWindow=SW_HIDE;
        si.hStdOutput=newstdout;
        si.hStdError=newstdout;
        if(lParam==0)
          {
            strcpy(outtxt,"Starting \"grubinst.exe");
            strcpy(&outtxt[22],&fn[strlen(sfn)]);
            strcpy(&outtxt[22+strlen(fn)-strlen(sfn)], "\"...\r\n\r\n");
          }
        if(!CreateProcess(NULL,fn,NULL,NULL,TRUE,CREATE_NEW_CONSOLE,NULL,NULL,&si,&pi))
          {
            CloseHandle(newstdout);
            CloseHandle(read_stdout);
            char errmsg[512];
            sprintf(errmsg,"%s\r\nError code: %d",LoadText(IDS_EXEC_ERROR),(int)GetLastError());
            MessageBox(hWnd,errmsg,NULL,MB_OK | MB_ICONERROR);
              {
                EndDialog(hOWnd,0);
                break;
              }
          }

        unsigned long exit=0;
        unsigned long bread;
        unsigned long avail;
      
        memset(buf,0,sizeof(buf));
        for(;;)
        {
          GetExitCodeProcess(pi.hProcess,&exit);
          if(exit !=STILL_ACTIVE)
            break;
          PeekNamedPipe(read_stdout,buf,1023,&bread,&avail,NULL);
          if(bread !=0)
          {
            memset(buf,0,sizeof(buf));
            if(avail > 1023)
            {
              while(bread >=1023)
              {
                ReadFile(read_stdout,buf,1023,&bread,NULL);
                strcat(outtxt,buf);
                memset(buf,0,sizeof(buf));
              }
            }
            else
            {
              ReadFile(read_stdout,buf,1023,&bread,NULL);
              strcat(outtxt,buf);
            }
          }
        }
        CloseHandle(pi.hThread);
        CloseHandle(pi.hProcess);
        CloseHandle(newstdout);
        CloseHandle(read_stdout);
      

        SETTEXTEX stx;
        stx.flags=ST_DEFAULT;
        stx.codepage=CP_ACP;
        SendDlgItemMessage(hOWnd,IDC_RICHEDIT,EM_SETTEXTEX,(WPARAM)&stx,(LPARAM)&outtxt);
    }
    else
    {
      MessageBox(hWnd,LoadText(IDS_NO_GRUBINST),NULL,MB_OK | MB_ICONERROR);
      EndDialog(hOWnd,0);
    }
    break;
  }
case WM_ACTIVATE:
    if(wParam==WA_ACTIVE)
       SendMessage(GetDlgItem(hOWnd,IDC_RICHEDIT),EM_SETSEL,(WPARAM)-1,(LPARAM)0);
    break;
  case WM_COMMAND:
    {
      switch(wParam & 0xFFFF)
      {
      case IDC_QUIT:
        {
          EndDialog(hOWnd,0);
          break;
        }
      case IDC_COPY:
        {
          SendDlgItemMessage(hOWnd, IDC_RICHEDIT, WM_COPY, 0,(LPARAM)0);
          break;
        }
      case IDC_SELALL:
        {
          SendMessage(GetDlgItem(hOWnd,IDC_RICHEDIT),EM_SETSEL,(WPARAM)0,(LPARAM)-1);
          break;
        }
      }
    }
  default:
    return FALSE;
  }
  return TRUE;
}


BOOL CALLBACK DialogMainProc(HWND hWnd,UINT wMsg,WPARAM wParam,LPARAM lParam)
{

  switch(wMsg){
  case WM_CLOSE:
    {
      EndDialog(hWnd,0);
      break;
    }
  case WM_INITDIALOG:
    {
      HICON hIcon;
      char title[30];

      ChangeText(hWnd);
      sprintf(title,"%s " VERSION,LoadText(IDS_MAIN));
      SendMessage(hWnd,WM_SETTEXT,0,(LPARAM)title);
      hIcon=LoadIcon(hInst,MAKEINTRESOURCE(ICO_MAIN));
      SendMessage(hWnd,WM_SETICON,ICON_BIG,(LPARAM)hIcon);
      SetDlgItemText(hWnd,IDC_TIMEOUT,"0");
      SendDlgItemMessage(hWnd,IDC_FILESYSTEM,CB_ADDSTRING,0,(LPARAM)"AutoDetect");
      SendDlgItemMessage(hWnd,IDC_FILESYSTEM,CB_ADDSTRING,0,(LPARAM)"MBR");
      SendDlgItemMessage(hWnd,IDC_FILESYSTEM,CB_ADDSTRING,0,(LPARAM)"NTFS");
      SendDlgItemMessage(hWnd,IDC_FILESYSTEM,CB_ADDSTRING,0,(LPARAM)"FAT32");
      SendDlgItemMessage(hWnd,IDC_FILESYSTEM,CB_ADDSTRING,0,(LPARAM)"FAT12/FAT16");
      SendDlgItemMessage(hWnd,IDC_FILESYSTEM,CB_ADDSTRING,0,(LPARAM)"EXT2/EXT3");
      SendDlgItemMessage(hWnd,IDC_FILESYSTEM,CB_SETCURSEL,(WPARAM)0,0);
      CheckDlgButton(hWnd,IDC_VERBOSE,BST_CHECKED);
      CheckDlgButton(hWnd,IDC_DISABLE_FLOPPY,BST_CHECKED);
      RefreshDisk(hWnd);
      break;
    }
  case WM_COMMAND:
    {
      switch(wParam & 0xFFFF){
      case IDC_TEST:
        if(LoadLibrary("riched20.dll"))
          DialogBoxParam(NULL,MAKEINTRESOURCE(DLG_OUTPUT),hWnd,DialogOutputProc,0);
        else
          Install(hWnd,TRUE);
        break;
      case IDC_INSTALL:
        if(LoadLibrary("riched20.dll"))
          DialogBoxParam(NULL,MAKEINTRESOURCE(DLG_OUTPUT),hWnd,DialogOutputProc,1);
        else
          Install(hWnd,FALSE);
        break;
      case IDC_QUIT:
        EndDialog(hWnd,0);
        break;
      case IDC_ISDISK:
      {
        if(wParam>>16==BN_CLICKED)
        {
          char temp[512]="";
          GetDlgItemText(hWnd,IDC_DISKS,temp,sizeof(temp));
          if(strcmp(temp,""))
            RefreshPart(hWnd,TRUE);
          else
            SendDlgItemMessage(hWnd,IDC_PARTS,CB_RESETCONTENT,0,0);
        }
        break;
      }
      case IDC_ISFILE:
      {
        if(wParam>>16==BN_CLICKED)
        {
          char temp[512]="";
          GetDlgItemText(hWnd,IDC_FILENAME,temp,sizeof(temp));
          if(strcmp(temp,""))
            RefreshPart(hWnd,TRUE);
          else
            SendDlgItemMessage(hWnd,IDC_PARTS,CB_RESETCONTENT,0,0);
        }
        break;
      }
      case IDC_REFRESH_DISK:
        RefreshDisk(hWnd);
        SendDlgItemMessage(hWnd,IDC_PARTS,CB_RESETCONTENT,0,0);
        break;
      case IDC_FILESYSTEM:
      {
        if(wParam>>16==CBN_SELENDOK)
        {
          char temp[512]="";
          if(IsDlgButtonChecked(hWnd,IDC_ISDISK)==BST_CHECKED)
            GetDlgItemText(hWnd,IDC_DISKS,temp,sizeof(temp));
          else
          if(IsDlgButtonChecked(hWnd,IDC_ISFILE)==BST_CHECKED)
            GetDlgItemText(hWnd,IDC_FILENAME,temp,sizeof(temp));
          if(strcmp(temp,""))
            RefreshPart(hWnd,TRUE);
        }
        break;
      }
      case IDC_REFRESH_PART:
        RefreshPart(hWnd,FALSE);
        break;
      case IDC_BROWSE:
        {
          OPENFILENAME ofn;
          char fn[MAX_PATH];

          memset(&ofn,0,sizeof(ofn));
          GetDlgItemText(hWnd,IDC_FILENAME,fn,sizeof(fn));
          ofn.lStructSize=sizeof(ofn);
          ofn.hwndOwner=hWnd;
          ofn.hInstance=hInst;
          ofn.lpstrFile=fn;
          ofn.nMaxFile=sizeof(fn);
          ofn.Flags=OFN_HIDEREADONLY | OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
          ofn.lpstrFilter="Raw disk image files (*.dsk, *.img, *.vhd)\0*.dsk;*.img;*.vhd\0All files (*.*)\0*.*\0\0";
          if(GetOpenFileName(&ofn))
          {
            SetDlgItemText(hWnd,IDC_FILENAME,fn);
            RefreshPart(hWnd,TRUE);
          }
          break;
        }
      case IDC_DISKS:
        {
          if(wParam>>16==CBN_SELENDOK)
            {
              CheckRadioButton(hWnd,IDC_ISDISK,IDC_ISFILE,IDC_ISDISK);
              RefreshPart(hWnd,TRUE);
            }
          break;
        }
      case IDC_RESTOREMBR:
        {
          if(wParam>>16==BN_CLICKED)
            {
              if(IsDlgButtonChecked(hWnd,IDC_RESTOREMBR)==BST_CHECKED)
              {
                CheckDlgButton(hWnd,IDC_SAVEMBR,BST_UNCHECKED);
                CheckDlgButton(hWnd,IDC_RESTORE_PREVMBR,BST_UNCHECKED);
                CheckDlgButton(hWnd,IDC_OUTPUT,BST_UNCHECKED);
                SetDlgItemText(hWnd,IDC_INSTALL,"Restore");
              }
              else
              {
                SetDlgItemText(hWnd,IDC_INSTALL,"Install");
              }
            }
          break;
        }
      case IDC_RESTORE_PREVMBR:
        {
          if(wParam>>16==BN_CLICKED)
            {
              if(IsDlgButtonChecked(hWnd,IDC_RESTORE_PREVMBR)==BST_CHECKED)
              {
                CheckDlgButton(hWnd,IDC_SAVEMBR,BST_UNCHECKED);
                CheckDlgButton(hWnd,IDC_RESTOREMBR,BST_UNCHECKED);
                CheckDlgButton(hWnd,IDC_OUTPUT,BST_UNCHECKED);
                SetDlgItemText(hWnd,IDC_INSTALL,"Restore");
              }
              else
                SetDlgItemText(hWnd,IDC_INSTALL,"Install");
            }
          break;
        }
      case IDC_SAVEMBR:
        {
          if(wParam>>16==BN_CLICKED)
            {
              if(IsDlgButtonChecked(hWnd,IDC_SAVEMBR)==BST_CHECKED)
              {
                CheckDlgButton(hWnd,IDC_RESTORE_PREVMBR,BST_UNCHECKED);
                CheckDlgButton(hWnd,IDC_RESTOREMBR,BST_UNCHECKED);
                CheckDlgButton(hWnd,IDC_OUTPUT,BST_UNCHECKED);
                SetDlgItemText(hWnd,IDC_INSTALL,"Save && Install");
              }
              else
                SetDlgItemText(hWnd,IDC_INSTALL,"Install");
            }
          break;
        }
      case IDC_OUTPUT:
        {
          if(wParam>>16==BN_CLICKED)
            {
              if(IsDlgButtonChecked(hWnd,IDC_OUTPUT)==BST_CHECKED)
              {
                CheckDlgButton(hWnd,IDC_RESTORE_PREVMBR,BST_UNCHECKED);
                CheckDlgButton(hWnd,IDC_RESTOREMBR,BST_UNCHECKED);
                CheckDlgButton(hWnd,IDC_SAVEMBR,BST_UNCHECKED);
                SetDlgItemText(hWnd,IDC_INSTALL,"Save");
              }
              else
                SetDlgItemText(hWnd,IDC_INSTALL,"Install");
            }
          break;
        }
      case IDC_FILENAME:
        {
          if(wParam>>16==EN_CHANGE)
            {
              CheckRadioButton(hWnd,IDC_ISDISK,IDC_ISFILE,IDC_ISFILE);
              SendDlgItemMessage(hWnd,IDC_PARTS,CB_RESETCONTENT,0,0);
            }
          break;
        }
      case IDC_BROWSE_SAVE:
        {
          OPENFILENAME ofn;
          char fn[512];

          memset(&ofn,0,sizeof(ofn));
          GetDlgItemText(hWnd,IDC_SAVEFILE,fn,sizeof(fn));
          ofn.lStructSize=sizeof(ofn);
          ofn.hwndOwner=hWnd;
          ofn.hInstance=hInst;
          ofn.lpstrFile=fn;
          ofn.nMaxFile=sizeof(fn);
          ofn.lpstrFilter="BIN files (*.bin)\0*.bin\0MBR files (*.mbr)\0*.mbr\0All files (*.*)\0*.*\0\0";
          ofn.Flags=OFN_HIDEREADONLY;
          if(IsDlgButtonChecked(hWnd,IDC_RESTOREMBR)==BST_CHECKED)
          {
            ofn.Flags|=OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;
            if(GetOpenFileName(&ofn))
              SetDlgItemText(hWnd,IDC_SAVEFILE,fn);
          }
          else
            if(IsDlgButtonChecked(hWnd,IDC_OUTPUT)==BST_CHECKED)
              {
                ofn.Flags|=OFN_OVERWRITEPROMPT;
                if(!strcmp(fn,""))
                {
                   strcpy(fn,"grldr.mbr");
                   ofn.nFilterIndex=2;
                }
                if(GetSaveFileName(&ofn))
                  SetDlgItemText(hWnd,IDC_SAVEFILE,fn);
              }
              else
              {
                ofn.Flags|=OFN_OVERWRITEPROMPT;
                if(!strcmp(fn,""))
                   strcpy(fn,"MBR.bin");
                if(GetSaveFileName(&ofn))
                {
                  SetDlgItemText(hWnd,IDC_SAVEFILE,fn);
                  CheckDlgButton(hWnd,IDC_SAVEMBR,BST_CHECKED);
                }
            }
          break;
        }
      }
      break;
    }
  default:
    return FALSE;
  }
  return TRUE;
}

int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,int nCmdShow)
{
  hInst=hInstance;
  if(!strncmp(lpCmdLine,"--lang=",7))
    lng_ext=&lpCmdLine[7];

  DialogBoxParam(hInst,MAKEINTRESOURCE(DLG_MAIN),NULL,DialogMainProc,0);
  return 0;
}

